"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.unfold = exports.foldWithIndex = exports.fold = exports.IndexedFunctor = exports.Functor = void 0;
var _registry = /*#__PURE__*/require("@traversable/registry");
var _core = /*#__PURE__*/require("./core.js");
const Functor = exports.Functor = {
  map(f) {
    return x => {
      switch (true) {
        default:
          return _registry.fn.exhaustive(x);
        case _core.t.isLeaf(x):
          return x;
        case x.tag === _registry.URI.eq:
          return _core.t.Eq.def(x.def);
        case x.tag === _registry.URI.array:
          return _core.t.Array.def(f(x.def));
        case x.tag === _registry.URI.record:
          return _core.t.Record.def(f(x.def));
        case x.tag === _registry.URI.optional:
          return _core.t.Optional.def(f(x.def));
        case x.tag === _registry.URI.tuple:
          return _core.t.Tuple.def(_registry.fn.map(x.def, f));
        case x.tag === _registry.URI.object:
          return _core.t.Object.def(_registry.fn.map(x.def, f));
        case x.tag === _registry.URI.union:
          return _core.t.Union.def(_registry.fn.map(x.def, f));
        case x.tag === _registry.URI.intersect:
          return _core.t.Intersect.def(_registry.fn.map(x.def, f));
      }
    };
  }
};
const IndexedFunctor = exports.IndexedFunctor = {
  ...Functor,
  mapWithIndex(f) {
    return (x, ix) => {
      switch (true) {
        default:
          return _registry.fn.exhaustive(x);
        case _core.t.isLeaf(x):
          return x;
        case x.tag === _registry.URI.eq:
          return _core.t.Eq.def(x.def);
        case x.tag === _registry.URI.array:
          return _core.t.Array.def(f(x.def, ix));
        case x.tag === _registry.URI.record:
          return _core.t.Record.def(f(x.def, ix));
        case x.tag === _registry.URI.optional:
          return _core.t.Optional.def(f(x.def, ix));
        case x.tag === _registry.URI.tuple:
          return _core.t.Tuple.def(_registry.fn.map(x.def, (y, iy) => f(y, [...ix, iy])));
        case x.tag === _registry.URI.object:
          return _core.t.Object.def(_registry.fn.map(x.def, (y, iy) => f(y, [...ix, iy])));
        case x.tag === _registry.URI.union:
          return _core.t.Union.def(_registry.fn.map(x.def, (y, iy) => f(y, [...ix, _registry.symbol.union, iy])));
        case x.tag === _registry.URI.intersect:
          return _core.t.Intersect.def(_registry.fn.map(x.def, (y, iy) => f(y, [...ix, _registry.symbol.intersect, iy])));
      }
    };
  }
};
const fold = exports.fold = /*#__PURE__*/_registry.fn.cata(Functor);
const unfold = exports.unfold = /*#__PURE__*/_registry.fn.ana(Functor);
const foldWithIndex_ = /*#__PURE__*/_registry.fn.cataIx(IndexedFunctor);
const foldWithIndex = algebra => (x, ix) => foldWithIndex_(algebra)(x, ix ?? []);
exports.foldWithIndex = foldWithIndex;
//# sourceMappingURL=functor.js.map