"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.is = void 0;
var _registry = /*#__PURE__*/require("@traversable/registry");
/** @internal */
const Array_isArray = globalThis.Array.isArray;
/** @internal */
const Object_keys = globalThis.Object.keys;
/** @internal */
const Object_entries = globalThis.Object.entries;
/** @internal */
const Object_hasOwnProperty = globalThis.Object.prototype.hasOwnProperty;
/** @internal */
const isComposite = u => !!u && typeof u === "object";
/** @internal */
function hasOwn(u, key) {
  return typeof key === "symbol" ? isComposite(u) && key in u : Object_hasOwnProperty.call(u, key);
}
const is = exports.is = {
  has: _registry.has,
  array,
  record,
  union,
  intersect,
  optional,
  object,
  tuple
};
function array(guard) {
  return u => Array_isArray(u) && u.every(guard);
}
function record(...args) {
  const [keyGuard, valueGuard] = args.length === 1 ? [() => true, args[0]] : args;
  return u => {
    return isComposite(u) && !Array_isArray(u) && Object_entries(u).every(([k, v]) => keyGuard(k) && valueGuard(v));
  };
}
function union(...qs) {
  return u => qs.some(q => q(u));
}
function intersect(...qs) {
  return u => qs.every(q => q(u));
}
function optional(guard) {
  return u => u === void 0 || guard(u);
}
function object(qs, $) {
  return u => {
    switch (true) {
      case !u:
        return false;
      case !isComposite(u):
        return false;
      case !$.treatArraysAsObjects && Array_isArray(u):
        return false;
      case $.optionalTreatment === 'exactOptional':
        return exactOptional(qs, u);
      case $.optionalTreatment === 'presentButUndefinedIsOK':
        return presentButUndefinedIsOK(qs, u);
      case $.optionalTreatment === 'treatUndefinedAndOptionalAsTheSame':
        return treatUndefinedAndOptionalAsTheSame(qs, u);
      default:
        throw globalThis.Error('               \
                                                      \
        ["@traversable/schema/predicates/object$"]    \
                                                      \
          Expected "optionalTreatment" to be one of:  \
                                                      \
            • "exactOptional"                         \
            • "presentButUndefinedIsOK"               \
            • "treatUndefinedAndOptionalAsTheSame"    \
                                                      \
          Got:                                        \
        ' + globalThis.JSON.stringify($.optionalTreatment));
    }
  };
}
function isOptionalSchema(u) {
  return !!u && u.tag === _registry.URI.optional;
}
function isRequiredSchema(u) {
  return !!u && !isOptionalSchema(u);
}
function isOptionalNotUndefinedSchema(u) {
  return !!u && isOptionalSchema(u) && u(undefined) === false;
}
function isUndefinedSchema(u) {
  return !!u && u[_registry.symbol.tag] === _registry.URI.undefined;
}
function exactOptional(qs, u) {
  for (const k in qs) {
    const q = qs[k];
    switch (true) {
      case isUndefinedSchema(q) && !hasOwn(u, k):
        return false;
      case isOptionalNotUndefinedSchema(q) && hasOwn(u, k) && u[k] === undefined:
        return false;
      case isOptionalSchema(q) && !hasOwn(u, k):
        continue;
      case !q(u[k]):
        return false;
      default:
        continue;
    }
  }
  return true;
}
function presentButUndefinedIsOK(qs, u) {
  for (const k in qs) {
    const q = qs[k];
    switch (true) {
      case isOptionalSchema(qs[k]) && !hasOwn(u, k):
        continue;
      case isOptionalSchema(qs[k]) && hasOwn(u, k) && u[k] === undefined:
        continue;
      case isOptionalSchema(qs[k]) && hasOwn(u, k) && q(u[k]):
        continue;
      case isOptionalSchema(qs[k]) && hasOwn(u, k) && !q(u[k]):
        return false;
      case isRequiredSchema(qs[k]) && !hasOwn(u, k):
        return false;
      case isRequiredSchema(qs[k]) && hasOwn(u, k) && q(u[k]) === true:
        continue;
      case hasOwn(u, k) && q(u[k]) === true:
        continue;
      default:
        return false;
    }
  }
  return true;
}
function treatUndefinedAndOptionalAsTheSame(qs, u) {
  const ixs = Object_keys(qs);
  for (const ix of ixs) {
    const q = qs[ix];
    const v = u[ix];
    if (!q(v)) return false;
  }
  return true;
}
function tuple(options) {
  return (...qs) => {
    const checkLength = xs => options?.minLength === void 0 ? xs.length === qs.length : options.minLength === -1 ? xs.length === qs.length : xs.length >= options.minLength && qs.length >= xs.length;
    return u => Array_isArray(u) && checkLength(u) && qs.every((q, ix) => q(u[ix]));
  };
}
//# sourceMappingURL=predicate.js.map