import type { Const, HKT, Identity, Kind, Mut, Mutable, TypeError } from '@traversable/registry';
import { URI } from '@traversable/registry';
import { Json } from '@traversable/json';
import type { SchemaOptions as Options } from './options.js';
import type { Guard, Predicate as AnyPredicate, TypePredicate, ValidateTuple } from './types.js';
import * as AST from './ast.js';
export type Source<T> = T extends (_: infer I) => unknown ? I : unknown;
export type Entry<S> = S extends {
    def: unknown;
} ? S : S extends Guard<infer T> ? t.Inline<T> : S extends (() => infer _ extends boolean) ? BoolLookup[`${_}`] : S;
export declare namespace Type {
    interface Array extends HKT {
        [-1]: this[0]['_type' & keyof this[0]][];
    }
    interface ReadonlyArray extends HKT {
        [-1]: readonly this[0]['_type' & keyof this[0]][];
    }
    interface Record extends HKT {
        [-1]: globalThis.Record<string, this[0]['_type' & keyof this[0]]>;
    }
    interface Optional extends HKT {
        [-1]: undefined | this[0]['_type' & keyof this[0]];
    }
    interface Object extends HKT {
        [-1]: Properties<this[0]>;
    }
    interface Tuple<LowerBound = t.Optional<any>> extends HKT {
        [-1]: Items<this[0], LowerBound>;
    }
    interface Intersect extends HKT {
        [-1]: Intersection<this[0]>;
    }
    interface Union extends HKT {
        [-1]: Unify<this[0]>;
    }
}
export type Predicate = AnyPredicate | Schema;
export type typeOf<T extends {
    _type?: unknown;
}, _ extends T['_type'] = T['_type']> = never | _;
export interface Unspecified extends LowerBound {
}
export interface LowerBound {
    (u: unknown): u is unknown;
    tag?: string;
    def?: unknown;
    _type?: unknown;
}
export interface Schema<Fn extends LowerBound = Unspecified> extends TypePredicate<Source<Fn>, Fn['_type']> {
    tag?: Fn['tag'];
    def?: Fn['def'];
    _type?: Fn['_type'];
}
export declare namespace t {
    interface Free extends HKT {
        [-1]: t.F<this[0]>;
    }
    type Fixpoint = t.Leaf | t.Eq.def<Json, Const> | t.Array.def<Fixpoint, Const> | t.Record.def<Fixpoint, Const> | t.Optional.def<Fixpoint, Const> | t.Object.def<{
        [x: string]: Fixpoint;
    }, Const> | t.Tuple.def<readonly Fixpoint[], t.Optional<any>, Const> | t.Union.def<readonly Fixpoint[], Const> | t.Intersect.def<readonly Fixpoint[], Const>;
    type F<_> = t.Leaf | t.Eq.def<_> | t.Array.def<_> | t.Optional.def<_> | t.Record.def<_> | t.Object.def<{
        [x: string]: _;
    }> | t.Tuple.def<readonly _[]> | t.Union.def<readonly _[]> | t.Intersect.def<readonly _[]>;
}
export declare namespace t {
    type _ = unknown;
    interface Inline<T> extends Guard<T> {
        readonly _type: T;
        tag: URI.inline;
    }
    interface Top {
        readonly _type: unknown;
        tag: URI.top;
    }
    interface Bottom {
        readonly _type: never;
        tag: URI.bottom;
    }
    interface InvalidSchema<_Err> extends TypeError<''>, t.Never {
    }
    interface Never extends Guard<never>, AST.never {
        readonly _type: never;
    }
    interface Any extends Guard<any>, AST.any {
        readonly _type: any;
    }
    interface Unknown extends Guard<_>, AST.unknown {
        readonly _type: unknown;
    }
    interface Void extends Guard<void>, AST.void {
        readonly _type: void;
    }
    interface Null extends Guard<null>, AST.null {
        readonly _type: null;
    }
    interface Undefined extends Guard<undefined>, AST.undefined {
        readonly _type: undefined;
    }
    interface BigInt extends Guard<bigint>, AST.bigint {
        readonly _type: bigint;
    }
    interface Symbol extends Guard<symbol>, AST.symbol {
        readonly _type: symbol;
    }
    interface Boolean extends Guard<boolean>, AST.boolean {
        readonly _type: boolean;
    }
    interface Integer extends Guard<number>, AST.integer {
        readonly _type: number;
    }
    interface Number extends Guard<number>, AST.number {
        readonly _type: number;
    }
    interface String extends Guard<string>, AST.string {
        readonly _type: string;
    }
}
export declare namespace t {
    const Never: t.Never;
    const Any: t.Any;
    const Unknown: t.Unknown;
    const Void: t.Void;
    const Null: t.Null;
    const Undefined: t.Undefined;
    const BigInt: t.BigInt;
    const Symbol: t.Symbol;
    const Boolean: t.Boolean;
    const Integer: t.Integer;
    const Number: t.Number;
    const String: t.String;
    function Inline<S>(guard: Guard<S>): t.Inline<S>;
    function Inline<P extends AnyPredicate>(guard: P): t.Inline<Entry<P>>;
    type Leaf = typeof Leaves[number];
    const Leaves: [Unknown, Never, Any, Void, Undefined, Null, Symbol, BigInt, Boolean, Integer, Number, String];
    const leafTags: ("@traversable/schema/URI::never" | "@traversable/schema/URI::unknown" | "@traversable/schema/URI::any" | "@traversable/schema/URI::void" | "@traversable/schema/URI::null" | "@traversable/schema/URI::undefined" | "@traversable/schema/URI::bigint" | "@traversable/schema/URI::symbol" | "@traversable/schema/URI::boolean" | "@traversable/schema/URI::integer" | "@traversable/schema/URI::number" | "@traversable/schema/URI::string")[];
    const isLeaf: (u: unknown) => u is Leaf;
    interface Eq<S = Unspecified> extends t.Eq.def<S> {
    }
    function Eq<V extends Mut<V>>(value: V, options?: Options): t.Eq<Mutable<V>>;
    function Eq<V extends Mut<V>>(value: V, options?: Options): t.Eq<Mutable<V>>;
    function Eq<const V>(value: V, options?: Options): t.Eq<Mutable<V>>;
    namespace Eq {
        interface def<T, F extends HKT = Identity> extends AST.eq<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<const T>(x: T, $?: Options): t.Eq.def<T>;
    }
    function Array<S extends Schema>(schema: S, readonly: 'readonly'): t.ReadonlyArray<S>;
    function Array<S extends Schema>(schema: S): t.Array<S>;
    interface Array<S extends Schema = Unspecified> extends t.Array.def<S> {
    }
    namespace Array {
        interface def<T, F extends HKT = Type.Array> extends AST.array<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T>(x: T): t.Array.def<T>;
    }
    interface ReadonlyArray<S extends Schema = Unspecified> extends t.Array.def<S, Type.ReadonlyArray> {
    }
    function Record<S extends Schema>(schema: S): t.Record<S>;
    interface Record<S extends LowerBound = Unspecified> extends t.Record.def<S> {
    }
    namespace Record {
        interface def<T, F extends HKT = Type.Record> extends AST.record<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T>(x: T): t.Record.def<T>;
    }
    function Union<S extends readonly Schema[]>(...schemas: S): t.Union<S>;
    interface Union<S extends readonly Schema[] = Schema[]> extends Union.def<S> {
    }
    namespace Union {
        interface def<T, F extends HKT = Type.Union> extends AST.union<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T extends readonly unknown[]>(xs: T): t.Union.def<T>;
    }
    function Intersect<S extends readonly Schema[]>(...schemas: S): t.Intersect<S>;
    interface Intersect<S extends readonly Schema[] = Schema[]> extends Intersect.def<S> {
    }
    namespace Intersect {
        interface def<T, F extends HKT = Type.Intersect> extends AST.intersect<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T extends readonly unknown[]>(xs: T): t.Intersect.def<T>;
    }
    function Optional<S extends Schema>(schema: S): t.Optional<S>;
    interface Optional<S extends Schema = Unspecified> extends t.Optional.def<S> {
    }
    namespace Optional {
        interface def<T, F extends HKT = Type.Optional> extends AST.optional<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T>(x: T): t.Optional.def<T>;
        const is: <S extends Schema>(u: unknown) => u is t.Optional<S>;
    }
    namespace Optional {
        interface Codec<S extends Schema = Unspecified, T extends Schema = Unspecified> extends TypePredicate<typeOf<S>, undefined | typeOf<T>> {
            decode(source: typeOf<S>): undefined | typeOf<T>;
            encode(target: undefined | typeOf<T>): typeOf<S>;
        }
    }
    function Object<S extends {
        [x: string]: Schema;
    }, T extends {
        [K in keyof S]: Entry<S[K]>;
    }>(schemas: S, options?: Options): t.Object<T>;
    function Object<S extends {
        [x: string]: Predicate;
    }, T extends {
        [K in keyof S]: Entry<S[K]>;
    }>(schemas: S, options?: Options): t.Object<T>;
    interface Object<S extends {
        [x: string]: unknown;
    } = {
        [x: string]: unknown;
    }> extends t.Object.def<S> {
    }
    namespace Object {
        interface def<T, F extends HKT = Type.Object> extends AST.object<T> {
            readonly _type: Kind<F, T>;
            readonly opt: Type.Optionals<T>[];
            (u: unknown): u is this['_type'];
        }
        function def<T extends {
            [x: string]: unknown;
        }>(ps: T, $?: Options): t.Object.def<T>;
    }
    function Tuple<S extends readonly Schema[], T extends {
        -readonly [Ix in keyof S]: Entry<S[Ix]>;
    }>(...schemas: ValidateTuple<S>): t.Tuple<Tuple.from<typeof schemas, T>>;
    function Tuple<S extends readonly Schema[], T extends {
        -readonly [Ix in keyof S]: Entry<S[Ix]>;
    }, V extends ValidateTuple<S>>(...schemas: [...guards: ValidateTuple<S>, options?: Options]): Tuple<Tuple.from<ValidateTuple<S>, T>>;
    interface Tuple<S extends readonly unknown[] = unknown[]> extends Tuple.def<S> {
    }
    namespace Tuple {
        interface def<T, LowerBound = t.Optional<any>, F extends HKT = Type.Tuple<LowerBound>> extends AST.tuple<T> {
            readonly _type: Kind<F, T>;
            (u: unknown): u is this['_type'];
        }
        function def<T extends readonly unknown[]>(xs: readonly [...T], $?: Options): t.Tuple.def<T>;
    }
    namespace Tuple {
        type InternalOptions = {
            minLength?: number;
        };
        type from<V extends readonly unknown[], T extends readonly unknown[]> = TypeError extends V[number] ? {
            [I in keyof V]: V[I] extends TypeError ? InvalidSchema<Extract<V[I], TypeError>> : V[I];
        } : T;
    }
}
//# sourceMappingURL=core.d.ts.map