import { URI, parseArgs } from '@traversable/registry';
import * as AST from './ast.js';
import { applyOptions, getConfig } from './config.js';
import { is as Combinator } from './predicate.js';
/** @internal */
const Object_assign = globalThis.Object.assign;
/** @internal */
const isPredicate = u => typeof u === 'function';
export var t;
(function (t) {
  t.Never = /*#__PURE__*/Object_assign(_ => false, AST.never);
  t.Any = /*#__PURE__*/Object_assign(_ => true, AST.any);
  t.Unknown = /*#__PURE__*/Object_assign(_ => true, AST.unknown);
  t.Void = /*#__PURE__*/Object_assign(_ => _ === void 0, AST.void);
  t.Null = /*#__PURE__*/Object_assign(_ => _ === null, AST.null);
  t.Undefined = /*#__PURE__*/Object_assign(_ => _ === void 0, AST.undefined);
  t.BigInt = /*#__PURE__*/Object_assign(_ => typeof _ === 'bigint', AST.bigint);
  t.Symbol = /*#__PURE__*/Object_assign(_ => typeof _ === 'symbol', AST.symbol);
  t.Boolean = /*#__PURE__*/Object_assign(_ => typeof _ === 'boolean', AST.boolean);
  t.Integer = /*#__PURE__*/Object_assign(globalThis.Number.isInteger, AST.integer);
  t.Number = /*#__PURE__*/Object_assign(_ => typeof _ === 'number', AST.number);
  t.String = /*#__PURE__*/Object_assign(_ => typeof _ === 'string', AST.string);
  function Inline(guard) {
    guard.tag = URI.inline;
    return guard;
  }
  t.Inline = Inline;
  t.Leaves = [t.Unknown, t.Never, t.Any, t.Void, t.Undefined, t.Null, t.Symbol, t.BigInt, t.Boolean, t.Integer, t.Number, t.String];
  t.leafTags = /*#__PURE__*/t.Leaves.map(_ => _.tag);
  t.isLeaf = u => typeof u === 'function' && 'tag' in u && typeof u.tag === 'string' && t.leafTags.includes(u.tag);
  function Eq(v, $) {
    return t.Eq.def(v, $);
  }
  t.Eq = Eq;
  (function (Eq) {
    function def(x, $ = getConfig().schema) {
      const config = applyOptions($);
      return Object_assign(src => typeof x === 'function' ? x(src) : config.eq.equalsFn(src, x), AST.eq(x));
    }
    Eq.def = def;
  })(Eq = t.Eq || (t.Eq = {}));
  function Array(x) {
    return t.Array.def(x);
  }
  t.Array = Array;
  (function (Array) {
    function def(x) {
      return Object_assign(src => isPredicate(x) ? Combinator.array(x)(src) : x, AST.array(x));
    }
    Array.def = def;
  })(Array = t.Array || (t.Array = {}));
  function Record(x) {
    return t.Record.def(x);
  }
  t.Record = Record;
  (function (Record) {
    function def(x) {
      return Object_assign(src => isPredicate(x) ? Combinator.record(x)(src) : x, AST.record(x));
    }
    Record.def = def;
  })(Record = t.Record || (t.Record = {}));
  function Union(...xs) {
    return t.Union.def(xs);
  }
  t.Union = Union;
  (function (Union) {
    function def(xs) {
      return Object_assign(src => xs.every(isPredicate) ? Combinator.union(...xs)(src) : xs, AST.union(xs));
    }
    Union.def = def;
  })(Union = t.Union || (t.Union = {}));
  function Intersect(...xs) {
    return t.Intersect.def(xs);
  }
  t.Intersect = Intersect;
  (function (Intersect) {
    function def(xs) {
      return Object_assign(src => xs.every(isPredicate) ? Combinator.intersect(...xs)(src) : xs, AST.intersect(xs));
    }
    Intersect.def = def;
  })(Intersect = t.Intersect || (t.Intersect = {}));
  function Optional(x) {
    return Optional.def(x);
  }
  t.Optional = Optional;
  (function (Optional) {
    function def(x) {
      return Object_assign(src => isPredicate(x) ? Combinator.optional(x)(src) : x, AST.optional(x));
    }
    Optional.def = def;
    Optional.is = u => !!u && (typeof u === 'function' || typeof u === 'object') && 'tag' in u && u.tag === URI.optional;
  })(Optional = t.Optional || (t.Optional = {}));
  function Object(xs, $ = getConfig().schema) {
    return t.Object.def(xs, $);
  }
  t.Object = Object;
  (function (Object) {
    function def(xs, $) {
      return Object_assign(src => Combinator.record(isPredicate)(xs) ? Combinator.object(xs, {
        ...getConfig().schema,
        ...$
      })(src) : xs, {
        opt: globalThis.Object.keys(xs).filter(k => Optional.is(xs[k]))
      }, AST.object(xs));
    }
    Object.def = def;
  })(Object = t.Object || (t.Object = {}));
  //
  function Tuple(...args) {
    const [guards, $] = parseArgs(getConfig().schema, args);
    return t.Tuple.def(guards, $);
  }
  t.Tuple = Tuple;
  (function (Tuple) {
    function def(xs, $ = getConfig().schema) {
      const options = {
        ...$,
        minLength: $.optionalTreatment === 'treatUndefinedAndOptionalAsTheSame' ? -1 : xs.findIndex(t.Optional.is)
      };
      return Object_assign(src => xs.every(isPredicate) ? Combinator.tuple(options)(...xs)(src) : xs, AST.tuple(xs));
    }
    Tuple.def = def;
  })(Tuple = t.Tuple || (t.Tuple = {}));
})(t || (t = {}));
//# sourceMappingURL=core.js.map