import { fn, URI, symbol } from '@traversable/registry';
import { t } from './core.js';
export const Functor = {
  map(f) {
    return x => {
      switch (true) {
        default:
          return fn.exhaustive(x);
        case t.isLeaf(x):
          return x;
        case x.tag === URI.eq:
          return t.Eq.def(x.def);
        case x.tag === URI.array:
          return t.Array.def(f(x.def));
        case x.tag === URI.record:
          return t.Record.def(f(x.def));
        case x.tag === URI.optional:
          return t.Optional.def(f(x.def));
        case x.tag === URI.tuple:
          return t.Tuple.def(fn.map(x.def, f));
        case x.tag === URI.object:
          return t.Object.def(fn.map(x.def, f));
        case x.tag === URI.union:
          return t.Union.def(fn.map(x.def, f));
        case x.tag === URI.intersect:
          return t.Intersect.def(fn.map(x.def, f));
      }
    };
  }
};
export const IndexedFunctor = {
  ...Functor,
  mapWithIndex(f) {
    return (x, ix) => {
      switch (true) {
        default:
          return fn.exhaustive(x);
        case t.isLeaf(x):
          return x;
        case x.tag === URI.eq:
          return t.Eq.def(x.def);
        case x.tag === URI.array:
          return t.Array.def(f(x.def, ix));
        case x.tag === URI.record:
          return t.Record.def(f(x.def, ix));
        case x.tag === URI.optional:
          return t.Optional.def(f(x.def, ix));
        case x.tag === URI.tuple:
          return t.Tuple.def(fn.map(x.def, (y, iy) => f(y, [...ix, iy])));
        case x.tag === URI.object:
          return t.Object.def(fn.map(x.def, (y, iy) => f(y, [...ix, iy])));
        case x.tag === URI.union:
          return t.Union.def(fn.map(x.def, (y, iy) => f(y, [...ix, symbol.union, iy])));
        case x.tag === URI.intersect:
          return t.Intersect.def(fn.map(x.def, (y, iy) => f(y, [...ix, symbol.intersect, iy])));
      }
    };
  }
};
export const fold = /*#__PURE__*/fn.cata(Functor);
export const unfold = /*#__PURE__*/fn.ana(Functor);
const foldWithIndex_ = /*#__PURE__*/fn.cataIx(IndexedFunctor);
export const foldWithIndex = algebra => (x, ix) => foldWithIndex_(algebra)(x, ix ?? []);
//# sourceMappingURL=functor.js.map